import codecs
from datetime import datetime
from glob import glob
import json
import os
import re
import warnings

import markdown2
import toml


class BaseSourceFile(object):
    def __init__(self, parent, filename):
        slug, _ = os.path.splitext(filename)
        self._source = filename
        self._slug = slug
        self._parent = parent
        self._config = self._read_configuration(filename)
        self._add_markdown_attachments()
        self._add_html_attachments()
        self._convert_markdown()

    @property
    def slug(self):
        return self._slug

    @property
    def url(self):
        # FIXME when _source_url is not set
        _url = self._parent._source_url
        _filter = {}
        for _arg in _url.arguments:
            try:
                _value = getattr(self, _arg)
                if type(_value) == list:
                    _filter[_arg] = _value[0]
                else:
                    _filter[_arg] = _value
            except AttributeError:
                warnings.warn(
                    'cannot create URL for "%s": no value for "%s"' % (
                        self._slug, _arg)
                )
                return None
        return _url.resolve(**_filter)

    @property
    def absolute_url(self):
        return '%s%s' % (self._parent.site_config['base_url'], self.url)

    def _read_configuration(self, filename):
        toml_file = '%s/%s' % (self._parent.source_dir, filename)
        with codecs.open(toml_file, encoding='utf-8') as configuration:
            return toml.loads(configuration.read())

    def _add_html_attachments(self):
        pattern = '%s/%s.*.html' % (self._parent.source_dir, self.slug)
        for attachment in glob(pattern):
            key = attachment.split('.')[-2]
            with codecs.open(attachment, encoding='utf-8') as content:
                if key in self._config:
                    warnings.warn(
                        '"%s" in %s overriden by HTML attachment.' % (
                            key, self.slug))
                self._config[key] = content.read()

    def _add_markdown_attachments(self):
        pattern = '%s/%s.*.markdown' % (self._parent.source_dir, self.slug)
        for attachment in glob(pattern):
            key = attachment.split('.')[-2] + '_markdown'
            with codecs.open(attachment, encoding='utf-8') as content:
                if key in self._config:
                    warnings.warn(
                        '"%s" in %s overriden by Markdown attachment.' % (
                            key, self.slug))
                self._config[key] = content.read()

    def _convert_markdown(self):
        add = {}
        for key in self._config:
            if key[-9:] == '_markdown':
                dest = key[:-9]
                add[dest] = markdown2.markdown(self._config[key])
                if dest in self._config:
                    warnings.warn(
                        '"%s" in %s overriden by Markdown conversion.' % (
                            dest, self.slug))
        self._config.update(add)

    def __getattr__(self, key):
        if key == 'slug':
            return self.slug
        if key in self._config:
            return self._config[key]
        if key in ['body', 'title']:
            # these are required for atom feeds, so if they've not been
            # set in the page's configuration, return them as empty
            return ''
        raise AttributeError

    def __getitem__(self, key):
        return self.__getattr__(key)

    def __iter__(self):
        _keys = self._config.keys()
        _keys.append('slug')
        return iter(_keys)

    def __repr__(self):
        return '<flourish.TomlSourceFile object (%s)>' % self._source

    class DoesNotExist(Exception):
        pass


class MarkdownSourceFile(BaseSourceFile):
    # FIXME, use re:
    # def split(self, text):
    #     FM_BOUNDARY = re.compile(r'^(?:{|})$', re.MULTILINE)
    #     _, fm, content = FM_BOUNDARY.split(text, 2)
    #     return "{" + fm + "}", content

    # FIXME use backticks for better viz on github (optionally '```toml')
    # +++ doesn't trigger the "nice" table like layout, but backticks
    # at least makes it presentable

    def _read_configuration(self, filename):
        markdown_file = '%s/%s' % (self._parent.source_dir, filename)
        with codecs.open(markdown_file, encoding='utf-8') as configuration:
            content = configuration.read()
        if content.startswith('---\n'):
            # 4 skips the starting `---\n`; 8 skips both
            end = content[4:].find('---\n')
            if end != -1:
                # we have a TOML block
                config = toml.loads(content[4:end+4])
                config['body_markdown'] = content[end+8:]
            else:
                raise RuntimeError(
                    '"%s" has no end marker for the frontmatter' % filename
                )
        else:
            config = {}
            config['body_markdown'] = content
        return config


class TomlSourceFile(BaseSourceFile):
    pass


class JsonSourceFile(BaseSourceFile):
    ISO8601 = re.compile(r'^\d{4}-\d{2}-\d{2}[ T]\d{2}:\d{2}:\d{2}Z$')

    def _read_configuration(self, filename):
        _json_file = '%s/%s' % (self._parent.source_dir, filename)
        with codecs.open(_json_file, encoding='utf-8') as _configuration:
            _config = json.loads(_configuration.read())

        for _key, _value in _config.iteritems():
            if type(_value) == unicode and self.ISO8601.match(_value):
                _config[_key] = datetime.strptime(_value, "%Y-%m-%dT%H:%M:%SZ")
        return _config



"""
The skeleton files that are created by running `flourish new <type>`
on the command-line.
"""

from textwrap import dedent


blog_files = {
    'source/_site.toml': dedent("""\
        author = 'Your Name Goes Here'
        base_url = 'http://example.com'
        title = 'Your New Flourish Blog'
    """),

    'source/generate.py': dedent("""\
        from flourish import helpers
        from flourish.generators import (
            IndexGenerator,
            PageGenerator,
        )


        class Homepage(IndexGenerator):
            order_by = ('published')
            template_name = 'homepage.html'


        def global_context(self):
            return {
                'dates': helpers.all_valid_dates(self.flourish),
            }

        GLOBAL_CONTEXT = global_context


        SOURCE_URL = (
            '/#slug',
            PageGenerator.as_generator(),
        )

        URLS = (
            (
                '/',
                'homepage',
                Homepage.as_generator()
            ),
            (
                '/#year/',
                'year-index',
                IndexGenerator.as_generator()
            ),
            (
                '/#year/#month',
                'month-index',
                IndexGenerator.as_generator()
            ),
        )
    """),

    'source/adding-pages.markdown': dedent("""\
        ---
        title = 'Adding pages'
        published = 2016-07-01T10:00:00Z
        ---

        Now let's add a new post to your blog. Using your text editor, create a
        new file called `learn-more-about-flourish.markdown` in the `source`
        directory, then copy and paste the text below into it:

            ---
            title = 'Learn more about Flourish'
            published = 2016-07-02T10:00:00Z
            ---

            This example site is a very brief introduction to using Flourish. For
            more information check out the [Flourish documentation][fd].

            If you find a problem, please file an [issue on GitHub][ish].

            To chat with other people who use Flourish, join [the Flourish Slack][slack].

            And you can follow [Flourish on twitter][tw] for updates.


            [fd]: http://withaflourish.net
            [ish]: https://github.com/norm/flourish/issues
            [slack]: http://slack.withaflourish.net
            [tw]: https://twitter.com/flourishpy

        Once you have saved this new post, click on the site header to return to the
        homepage. You'll notice that a new link has appeared ("Learn more about
        Flourish"). Click on this new post to continue.
    """),   # noqa: E501

    'source/editing-the-site.markdown': dedent("""\
        ---
        title = 'Editing the site'
        published = 2016-06-30T10:00:00Z
        tag = 'editing'
        ---

        To learn more about editing a Flourish powered web site, first open the
        directory you ran `flourish example` in into your text editor. Open the file
        `editing-the-site.markdown` in the `source` directory. This file contains
        Markdown that Flourish uses to create the web page you are reading.

        Add, change or delete some of the words in this paragraph, save the file
        and reload the page in your browser to see the changes take effect. When you
        are creating the content for your new Flourish site, this is how you can
        preview it as you write.

        ## Changing the site-wide configuration

        There is a file `_site.toml` in the `source` directory. It is not the source
        for any specific page, but rather it is site-wide configuration. 

        Open it in your text editor and change the text "Your Name Goes Here" to
        contain your name, save the file, reload this webpage and scroll to the
        bottom. You'll see your name appear in the copyright statement in the footer.

        Now you have a flavour of how to see your changes take effect,
        let's [add a new page](/adding-pages).
    """),   # noqa: E501

    'source/welcome.markdown': dedent("""\
        ---
        title = 'Welcome to your new blog'
        published = 2016-06-29T10:00:00Z
        ---

        Welcome to your new blog, powered by Flourish. This example content will 
        walk you through the basics of getting started using Flourish.

        Try clicking around the site. See how the title at the top of the page always
        takes you to the homepage and that there are three posts linked from there;
        that "2016" takes you to a similar list of pages, but the "06" and "07" links
        show less links? These links are all programmatically generated from the 
        site's source.

        The homepage links to all pages, the "2016" page links to all pages published
        in the year 2016, and "06" and "07" link to pages published in June and July
        of that year.

        Having familiarised yourself with the site as it is now, let's start 
        [editing it!](/editing-the-site)
    """),   # noqa: E501

    'source/site.css': dedent("""\
        body {
            max-width: 600px;
            margin: 2em auto;
            font-size: 1.2em;
            line-height: 1.4;
        }

        header {
            margin: 0 -40px 0;
            padding: 0 40px;
            border-bottom: 1px solid #ccc;
        }
        header h1,
        header p {
            font-size: 1.25em;
            font-weight: bold;
        }

        header nav {
            padding: 0;
            margin: 0 0 1em;
        }
        header nav ul, header nav li {
            padding: 0;
            margin: 0;
            list-style: none;
        }
        header nav li ul, header nav li li {
            display: inline-block;
        }
        header nav ul ul li:after {
            content: ', ';
        }
        header nav ul ul li:last-child:after {
            content: '';
        }

        #posts {
            margin: 2em 0;
            padding: 0;
        }
        #posts li {
            margin: 0 0 2em;
            padding: 0;
            list-style: none;
        }
        #posts li time {
            color: #666;
            display: block;
        }
        #posts li a {
            font-size: 1.5em;
        }

        pre {
            margin: 0 1em;
            padding: 0.5em 1em;
            overflow: scroll;
            background: #dee;
            border: 1px solid #599;
        }
        code {
            color: #266;
            background: #dee;
        }
        p > code {
            border: 1px solid #599;
            display: inline-block;
            padding: 1px 3px;
            line-height: 1.2;
        }

        footer {
            margin: 4em -40px 2em;
            padding: 0 40px;
            border-top: 1px solid #ccc;
        }
    """),

    'templates/base.html': dedent("""\
        <html lang='en-GB'>
        <head>
          <meta charset='UTF-8'>
          <title>{{title}}</title>
          <link rel='stylesheet' href='/site.css'>
        </head>
        <body>

        <header>
          {% block header %}
            <p>
              <a href='/'>
               {{site.title}}
              </a>
            </p>
          {% endblock %}

          <nav>
            <ul>
              {% for year in global.dates %}
                <li>
                  <a href='/{{year.year}}/'>{{year.year}}</a>:
                  <ul>
                    {% for month in year.months %}
                      <li><a href='/{{year.year}}/{{month.month}}'>{{month.month}}</a></li>
                    {% endfor %}
                  </ul>
                </li>
              {% endfor %}
            </ul>
          </nav>
        </header>


        <div id='body'>
          {% block body %}{% endblock %}
        </div>

        <footer>
          <p>
            Copyright &copy; 2016 {{site.author}}.
            Site made <a href='http://withaflourish.net'>with a flourish</a>.
          </p>
        </footer>

        </body>
        </html>
    """),   # noqa: E501

    'templates/homepage.html': dedent("""\
        {% extends "index.html" %}

        {% block header %}
          <h1>{{site.title}}</h1>
        {% endblock %}
    """),

    'templates/index.html': dedent("""\
        {% extends "base.html" %}

        {% block body %}

          <ul id='posts'>
            {% for page in pages %}
              <li>
                {% if page.published %}
                  <time datetime='{{page.published.strftime("%Y-%m-%dT%H:%M:%SZ")}}'>
                    {{page.published.strftime('%A %B %d, %Y')}}
                  </time>
                {% endif %}
                <a href='{{page.url}}'>{{page.title}}</a>
              </li>
            {% endfor %}
          </ul>

        {% endblock %}
    """),   # noqa: E501

    'templates/page.html': dedent("""\
        {% extends "base.html" %}

        {% block body %}

          <h1>{{page.title}}</h1>
          {{page.body}}

        {% endblock %}
    """),
}



import os


def relative_list_of_files_in_directory(directory):
    _file_list = []
    _trim = len(directory) + 1
    for _root, _dirs, _files in os.walk(directory):
        _subdir = _root[_trim:]
        for _file in _files:
            if len(_subdir):
                _file_list.append('%s/%s' % (_subdir, _file))
            else:
                _file_list.append(_file)
    return _file_list



# encoding: utf8

from datetime import datetime
from imp import load_source
from operator import attrgetter
import os
from shutil import copyfile, rmtree
import warnings

from jinja2 import Environment, FileSystemLoader
import toml

from .source import JsonSourceFile, MarkdownSourceFile, TomlSourceFile
from .url import URL
from .version import __version__    # noqa: F401


class Flourish(object):
    ARGS = [
        'source_dir',
        'templates_dir',
        'jinja',
    ]
    DATA = [
        '_assets',
        '_cache',
        '_filters',
        '_order_by',
        '_slice',
        '_source_files',
        '_source_url',
        '_urls',
    ]

    def __init__(
        self,
        source_dir='source',
        templates_dir='templates',
        output_dir='output',
        global_context=None,
        **kwargs
    ):
        self.source_dir = source_dir
        self.templates_dir = templates_dir
        self.output_dir = output_dir
        self.global_context = global_context
        self.jinja = Environment(
            loader=FileSystemLoader(self.templates_dir),
            keep_trailing_newline=True,
        )
        self.jinja.globals['url'] = self.template_resolve_url

        self._assets = []
        self._cache = []
        self._filters = []
        self._order_by = []
        self._slice = None
        self._source_files = []
        self._source_url = None
        self._urls = {}

        if not os.path.isdir(self.source_dir):
            raise AttributeError(
                'source_dir "%s" must exist' % self.source_dir)

        for _opt in self.DATA:
            if _opt in kwargs:
                if kwargs[_opt] is not None:
                    setattr(self, _opt, kwargs[_opt])

        self.site_config = self._read_site_config()

        if '_source_files' not in kwargs:
            self._add_sources()

        generate = load_source('generate', '%s/generate.py' % self.source_dir)
        try:
            self.set_global_context(getattr(generate, 'GLOBAL_CONTEXT'))
        except AttributeError:
            # global context is optional
            pass

        has_urls = False
        try:
            self.canonical_source_url(*generate.SOURCE_URL)
            has_urls = True
        except AttributeError:
            # source URLs are optional
            pass

        try:
            for url in generate.URLS:
                has_urls = True
                self.add_url(*url)
        except:
            # other URLs are optional
            pass

        if not has_urls:
            warnings.warn('There are no URLs configured in generate.py')

    @property
    def sources(self):
        return self

    def all(self):
        """ Get all source documents. """
        return self

    def count(self):
        """ Get the count of all source documents. """
        _sources = 0
        for source in self.sources.all():
            _sources += 1
        return _sources

    def get(self, slug):
        """ Get a single source document by slug. """
        for source in self._cache:
            if source.slug == slug:
                return source
        raise TomlSourceFile.DoesNotExist

    def filter(self, **kwargs):
        _clone = self.clone()
        for _key, _value in kwargs.iteritems():
            _clone._filters.append((_key, _value))
        return _clone

    def exclude(self, **kwargs):
        _clone = self.clone()
        for _key, _value in kwargs.iteritems():
            if '__' in _key:
                _field, _operator = _key.split('__', 2)
            else:
                _field, _operator = _key, 'eq'
            _new_operator = INVERSE_OPERATORS.get(_operator)
            _key = '%s__%s' % (_field, _new_operator)
            _clone._filters.append((_key, _value))
        return _clone

    def order_by(self, *args):
        return self.clone(_order_by=args)

    def canonical_source_url(self, url, generator):
        # FIXME hmmmm
        self._source_url = URL(self, url, 'source', generator)
        self.add_url(url, 'source', generator)

    def add_url(self, url, name, generator):
        _url_dict = {name: URL(self, url, name, generator)}
        self._urls.update(_url_dict)

    def resolve_url(self, name, **kwargs):
        return self._urls[name].resolve(**kwargs)

    def template_resolve_url(self, name, **kwargs):
        try:
            return self.resolve_url(name, **kwargs)
        except KeyError:
            warnings.warn(
                'Cannot resolve URL "%s", it has missing arguments' % name
            )
            return ''

    def all_valid_filters_for_url(self, name):
        return self._urls[name].all_valid_filters()

    def generate_site(self, report=False):
        if os.path.exists(self.output_dir):
            rmtree(self.output_dir)
        os.makedirs(self.output_dir)
        self.generate_all_urls(report=report)
        self.copy_assets(report=report)

    def generate_all_urls(self, report=False):
        for _entry in self._urls:
            self.generate_url(_entry, report=report)

    def generate_url(self, name, report=False):
        url = self._urls[name]
        url.generator(self, url, self.global_context, report=report)

    def set_global_context(self, global_context):
        self.global_context = global_context

    def copy_assets(self, report=False):
        for _file in self._assets:
            _source = '%s/%s' % (self.source_dir, _file)
            _destination = '%s/%s' % (self.output_dir, _file)
            _directory = os.path.dirname(_destination)
            if not os.path.isdir(_directory):
                os.makedirs(_directory)
            copyfile(_source, _destination)
            if report:
                print '->', _destination

    def clone(self, **kwargs):
        for _option in self.ARGS + self.DATA:
            _value = getattr(self, _option)
            if type(_value) == list:
                kwargs.setdefault(_option, _value[:])
            else:
                kwargs.setdefault(_option, _value)
        return type(self)(**kwargs)

    def _read_site_config(self):
        _config_file = '%s/_site.toml' % self.source_dir
        with open(_config_file) as _file:
            _config = toml.loads(_file.read())
        for key in ('author', 'title', 'base_url'):
            if key not in _config:
                raise RuntimeError(
                    '"%s" is a required entry in _site.toml' % key)
        return _config

    def _add_sources(self):
        """ Find source documents and register them. """
        trim = len(self.source_dir) + 1
        for root, dirs, files in os.walk(self.source_dir):
            this_dir = root[trim:]
            for file in files:
                # FIXME check for slug-ishness and otherwise ignore
                # (this could simplify _site.toml by being just another
                # ignored filename?)
                is_attachment_file = (
                    file.endswith(('.markdown', '.html')) and
                    len(file.split('.')) == 3
                )

                if len(this_dir):
                    file = '%s/%s' % (this_dir, file)

                if file == '_site.toml':
                    continue
                if file.startswith('generate.py'):
                    continue

                if file.endswith('.toml'):
                    self._cache.append(TomlSourceFile(self, file))
                elif file.endswith('.markdown') and len(file.split('.')) == 2:
                    self._cache.append(MarkdownSourceFile(self, file))
                elif file.endswith('.json'):
                    self._cache.append(JsonSourceFile(self, file))
                elif not is_attachment_file:
                    self._assets.append(file)
        self._source_files = list(self._cache)

    def get_valid_filters_for_tokens(self, tokens, objects=None):
        if objects is None:
            objects = self.sources

        _first_token = tokens[0]

        _values = set()
        for _source in objects:
            _is_date_filter = (
                _first_token in ['year', 'month', 'day'] and
                'published' in _source and
                type(_source['published'] == datetime)
            )

            if _is_date_filter:
                _value = getattr(_source['published'], _first_token)
                _values.add('%02d' % _value)
            elif _first_token in _source:
                if type(_source[_first_token]) == list:
                    for _value in _source[_first_token]:
                        _values.add(_value)
                else:
                    _values.add(_source[_first_token])

        _filters = []
        for _value in _values:
            _dict = {_first_token: _value}
            if len(tokens) == 1:
                _filters.append(_dict)
            else:
                _sub_objects = objects.filter(**_dict)
                _sub_tokens = self.get_valid_filters_for_tokens(
                    tokens[1:], _sub_objects)
                for _sub_token in _sub_tokens:
                    _update = _dict.copy()
                    _update.update(_sub_token)
                    _filters.append(_update)

        # the sort is only so the tests can compare results easily
        return sorted(_filters)

    def __len__(self):
        return self.count()

    def __getitem__(self, item):
        if isinstance(item, slice):
            if (item.start < 0) or (item.stop < 0):
                raise ValueError('Cannot use negative indexes with Flourish')
            return self.clone(_slice=item)
        if item < 0:
            raise ValueError('Cannot use negative indexes with Flourish')
        _iterator = iter(self)
        try:
            for i in range(0, item+1):
                obj = next(_iterator)
        except StopIteration:
            raise IndexError(item)
        return obj
        return None

    def _get_filtered_sources(self):
        _sources = []
        for _source in self._source_files:
            _add = True
            if len(self._filters):
                for _filter in self._filters:
                    if '__' in _filter[0]:
                        _field, _operator = _filter[0].split('__', 2)
                    else:
                        _field, _operator = _filter[0], 'eq'
                    _operation = OPERATORS.get(_operator)
                    if _operation is None:
                        # FIXME write test and raise more useful error
                        raise RuntimeError
                    _is_date_filter = (
                        _field in ['year', 'month', 'day'] and
                        'published' in _source and
                        type(_source['published'] == datetime)
                    )

                    try:
                        if _is_date_filter:
                            _test = '%02d' % getattr(
                                _source['published'], _field)
                        else:
                            _test = getattr(_source, _field)
                    except AttributeError:
                        _test = None

                    _result = _operation(_test, _filter[1])
                    if not _result:
                        _add = False
            if _add:
                _sources.append(_source)
        if self._slice is not None:
            _sources = _sources.__getitem__(self._slice)
        return _sources

    def __iter__(self):
        _sources = self._get_filtered_sources()
        for _order in self._order_by:
            if _order[0] == '-':
                _rev = True
                _attr = _order[1:]
            else:
                _rev = False
                _attr = _order
            try:
                _sources = sorted(
                    _sources, key=attrgetter(_attr), reverse=_rev)
            except AttributeError:
                warnings.warn(
                    'sorting sources by "%s" failed: '
                    'not all sources have that attribute' % _order
                )

        return iter(_sources)

    def __repr__(self):
        return '<flourish.Flourish object (source=%s)' % self.source_dir


def _equal_or_inside(value, test):
    if type(value) == list:
        return test in value
    else:
        return value == test


def _not_equal_or_inside(value, test):
    return not _equal_or_inside(value, test)


def _less_than(value, test):
    return value < test


def _less_than_or_equal_to(value, test):
    return value <= test


def _greater_than(value, test):
    return value > test


def _greater_than_or_equal_to(value, test):
    return value >= test


def _contains(value, test):
    try:
        if type(value) == list:
            for _v in value:
                if test in _v:
                    return True
            return False
        else:
            return test in value
    except TypeError:
        return False


def _excludes(value, test):
    return not _contains(value, test)


def _inside(value, test):
    return value in test


def _outside(value, test):
    return value not in test


def _set(value, test):
    return value is not None


def _unset(value, test):
    return value is None


OPERATORS = {
    'eq': _equal_or_inside,
    'neq': _not_equal_or_inside,
    'lt': _less_than,
    'lte': _less_than_or_equal_to,
    'gt': _greater_than,
    'gte': _greater_than_or_equal_to,
    'contains': _contains,
    'excludes': _excludes,
    'in': _inside,
    'notin': _outside,
    'set': _set,
    'unset': _unset,
}
INVERSE_OPERATORS = {
    'neq': 'eq',
    'eq': 'neq',
    'lt': 'gte',
    'lte': 'gt',
    'gt': 'lte',
    'gte': 'lt',
    'contains': 'excludes',
    'excludes': 'contains',
    'in': 'notin',
    'notin': 'in',
    'set': 'unset',
    'unset': 'set',
}



from collections import Sequence
from math import ceil


# FIXME add in concertina nav
class Paginator(object):
    def __init__(self, object_list=[], per_page=10, base_url='/'):
        self.object_list = list(object_list)
        self.per_page = per_page
        self.base_url = base_url

    @property
    def count(self):
        return len(self.object_list)

    @property
    def num_pages(self):
        if self.count == 0:
            return 1
        else:
            return int(ceil(len(self.object_list) / float(self.per_page)))

    @property
    def page_range(self):
        return range(1, self.num_pages + 1)

    def page(self, number):
        # FIXME check this is a valid page
        start_index = (number - 1) * self.per_page
        end_index = start_index + self.per_page
        return Page(self.object_list[start_index:end_index], number, self)

    def pages(self):
        _pages = []
        for _page in self.page_range:
            _pages.append(self.page(_page))
        return _pages

    def __iter__(self):
        return iter(self.pages())


class NoPage(Exception):
    pass


class Page(Sequence):
    def __init__(self, object_list, number, paginator):
        self.object_list = object_list
        self.number = number
        self.paginator = paginator

    def __len__(self):
        return len(self.object_list)

    def __getitem__(self, index):
        return self.object_list[index]

    def has_next(self):
        return self.number < self.paginator.num_pages

    def next_page_number(self):
        if self.has_next():
            return self.number + 1
        else:
            raise NoPage

    def has_previous(self):
        return self.number > 1

    def previous_page_number(self):
        if self.has_previous():
            return self.number - 1
        else:
            raise NoPage

    def has_other_pages(self):
        return self.has_next() or self.has_previous()

    def start_index(self):
        return ((self.number - 1) * self.paginator.per_page) + 1

    def end_index(self):
        return self.number * self.paginator.per_page

    @property
    def url(self):
        if self.number > 1:
            return '%s%s' % (self.paginator.base_url, 'page-%s' % self.number)
        else:
            return self.paginator.base_url



# encoding: utf8

from collections import defaultdict


def all_valid_dates(flourish, key='published'):
    def recursively_default_dict():
        return defaultdict(recursively_default_dict)

    _captured = recursively_default_dict()
    for _source in flourish.sources.all():
        try:
            _date = getattr(_source, key)
            _captured[_date.year][_date.month][_date.day] = 1
        except AttributeError:
            pass

    _dates = []
    for _year in sorted(_captured):
        _months = []
        for _month in sorted(_captured[_year]):
            _days = []
            for _day in sorted(_captured[_year][_month]):
                _days.append({'day': '%02d' % _day})
            _months.append({'month': '%02d' % _month, 'days': _days})
        _dates.append({'year': '%04d' % _year, 'months': _months})
    return _dates


def publication_range(flourish, key='published'):
    _lowest = None
    _highest = None

    for _source in flourish.sources.all():
        try:
            _pub = getattr(_source, key)
            _year = _pub.year
            if _lowest is None or _year < _lowest:
                _lowest = _year
            if _highest is None or _year > _highest:
                _highest = _year
        except AttributeError:
            pass

    if _lowest is not None and _highest is not None:
        return u'%d–%d' % (_lowest, _highest)



from setuptools import find_packages, setup

# read in __version__
execfile('flourish/version.py')


setup(
    name='flourish',
    version=__version__,    # noqa: F821

    description='Static website generator',
    # long_description='',
    url='https://github.com/norm/flourish',

    license='MIT',
    author='Mark Norman Francis',
    author_email='norm@201created.com',

    packages=find_packages(exclude=['tests']),
    entry_points={
        'console_scripts': ['flourish=flourish.command_line:main'],
    },

    install_requires=[
        'boto3',
        'Flask',
        'Jinja2',
        'markdown2',
        'pyatom',
        'toml.py',
        'watchdog',
    ],

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 2',
        # 'Programming Language :: Python :: 3',

        'Topic :: Software Development :: Build Tools',
        'Topic :: Software Development :: Code Generators',
        'Topic :: Text Processing :: Markup :: HTML',
        'Topic :: Internet :: WWW/HTTP :: Site Management',
    ],


    # # What does your project relate to?
    # keywords='sample setuptools development',

    # # You can just specify the packages manually here if your project is
    # # simple. Or you can use find_packages().
    # packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    # # Alternatively, if you want to distribute just a my_module.py, uncomment
    # # this:
    # #   py_modules=["my_module"],

    # # List run-time dependencies here.  These will be installed by pip when
    # # your project is installed. For an analysis of "install_requires" vs pip's
    # # requirements files see:
    # # https://packaging.python.org/en/latest/requirements.html
    # install_requires=['peppercorn'],

    # # List additional groups of dependencies here (e.g. development
    # # dependencies). You can install these using the following syntax,
    # # for example:
    # # $ pip install -e .[dev,test]
    # extras_require={
    #     'dev': ['check-manifest'],
    #     'test': ['coverage'],
    # },

    # # If there are data files included in your packages that need to be
    # # installed, specify them here.  If using Python 2.6 or less, then these
    # # have to be included in MANIFEST.in as well.
    # package_data={
    #     'sample': ['package_data.dat'],
    # },

    # # Although 'package_data' is the preferred approach, in some case you may
    # # need to place data files outside of your packages. See:
    # # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    # data_files=[('my_data', ['data/data_file'])],

    # # To provide executable scripts, use entry points in preference to the
    # # "scripts" keyword. Entry points provide cross-platform support and allow
    # # pip to create the appropriate form of executable for the target platform.
    # entry_points={
    #     'console_scripts': [
    #         'sample=sample:main',
    #     ],
    # },
)



from datetime import datetime
import os

from pyatom import AtomFeed

from .paginator import Paginator


class MissingValue(Exception):
    pass


class PageContextMixin(object):
    def get_context_data(self):
        _context = super(PageContextMixin, self).get_context_data()
        _context['page'] = self.source_objects[0]
        _context.update(self.source_objects[0]._config)
        return _context


class PageIndexContextMixin(object):
    def get_context_data(self):
        _context = super(PageIndexContextMixin, self).get_context_data()
        _context['pages'] = self.source_objects
        return _context


class BaseGenerator(object):
    order_by = None
    sources_exclude = None
    sources_filter = None
    template_name = None
    limit = None

    @classmethod
    def as_generator(cls):
        def generator(flourish, url, global_context=None, report=False):
            self = cls(flourish, url, global_context, report)
            return self.generate()

        return generator

    def __init__(self, flourish, url, global_context, report):
        self.flourish = flourish
        self.url = url
        self.global_context = global_context
        self.current_url = None
        self.report = report

    def generate(self):
        for _tokens in self.get_url_tokens():
            self.get_current_url(_tokens)
            self.get_objects(_tokens)
            self.output_to_file()

    def get_url_tokens(self):
        self.valid_filters = \
            self.flourish.all_valid_filters_for_url(self.url.name)
        return self.valid_filters

    def get_current_url(self, tokens):
        self.current_url = self.url.resolve(**tokens)
        return self.current_url

    def get_objects(self, tokens):
        _filtered = self.get_filtered_sources().filter(**tokens)
        _ordering = self.get_order_by()
        if _ordering is not None:
            _filtered = _filtered.order_by(self.order_by)

        if self.limit is not None:
            self.source_objects = _filtered[0:self.limit]
        else:
            self.source_objects = _filtered
        return self.source_objects

    def get_filtered_sources(self):
        _sources = self.flourish.sources
        if self.sources_filter is not None:
            _sources = self.flourish.filter(**self.sources_filter)
        if self.sources_exclude is not None:
            _sources = self.flourish.exclude(**self.sources_exclude)
        return _sources

    def get_order_by(self):
        return self.order_by

    def output_to_file(self):
        _filename = self.get_output_filename()
        if self.report:
            print '->', _filename

        _rendered = self.render_output()
        _directory = os.path.dirname(_filename)
        if not os.path.isdir(_directory):
            os.makedirs(_directory)
        with open(_filename, 'w') as _output:
            _output.write(_rendered.encode('utf-8'))

    def get_output_filename(self):
        _destination = '%s%s' % (self.flourish.output_dir, self.current_url)
        if _destination.endswith('/'):
            _destination = _destination + 'index.html'
        if not _destination.endswith(('.html', '.atom')):
            _destination = _destination + '.html'
        return _destination

    def render_output(self):
        _context = self.get_context_data()
        _template = self.get_template()
        return self.render_template(_template, _context)

    def get_context_data(self):
        _context = {}
        _context['objects'] = self.source_objects
        _context['site'] = self.flourish.site_config
        _context['current_url'] = self.current_url
        if self.global_context is not None:
            _context['global'] = self.global_context(self)
        return _context

    def get_template(self):
        _name = self.get_template_name()
        if _name is None:
            raise MissingValue
        return self.flourish.jinja.get_template(_name)

    def get_template_name(self):
        return self.template_name

    def render_template(self, template, context_data):
        return template.render(context_data)


class PageGenerator(PageContextMixin, BaseGenerator):
    template_name = 'page.html'

    def get_template_name(self):
        if 'template' in self.source_objects[0]:
            return self.source_objects[0]['template']
        if 'type' in self.source_objects[0]:
            return '%s.html' % self.source_objects[0]['type']
        return super(PageGenerator, self).get_template_name()


class IndexGenerator(PageIndexContextMixin, BaseGenerator):
    template_name = 'index.html'


class PaginatedIndexGenerator(IndexGenerator):
    per_page = 10

    def generate(self):
        for _tokens in self.get_url_tokens():
            _base_url = self.get_current_url(_tokens)
            _paginator = self.get_paginated_objects(_tokens, _base_url)
            for _page in _paginator:
                self.current_url = _page.url
                self.source_objects = _page.object_list
                self.current_page = _page
                self.output_to_file()

    def get_paginated_objects(self, tokens, base_url):
        _objects = self.get_objects(tokens)
        _per_page = self.get_per_page()
        self.paginator = Paginator(_objects, _per_page, base_url)
        return self.paginator

    def get_per_page(self):
        return self.per_page

    def get_context_data(self):
        _context = super(PaginatedIndexGenerator, self).get_context_data()
        _context['current_page'] = self.current_page
        _context['pagination'] = self.paginator
        return _context


class AtomGenerator(BaseGenerator):
    order_by = ('-published')

    def get_objects(self, tokens):
        """
        Only consider objects that have the key "published" with a
        datetime value that is in the past.
        """
        _now = datetime.now()
        _sources = self.get_filtered_sources()
        _already_published = _sources.filter(published__lt=_now)
        _filtered = _already_published.filter(**tokens)
        _ordered = _filtered.order_by(self.order_by)

        if self.limit is not None:
            self.source_objects = _ordered[0:self.limit]
        else:
            self.source_objects = _ordered

        return self.source_objects

    def render_output(self):
        _feed_global = {
            'author': self.flourish.site_config['author'],
            'title': self.flourish.site_config['title'],
            'url': self.flourish.site_config['base_url'],
            'feed_url': '%s%s' % (
                self.flourish.site_config['base_url'],
                self.current_url,
            ),
        }
        _feed = AtomFeed(**_feed_global)

        for _object in self.source_objects:
            entry = {
                'title': _object.title,
                'content': _object.body,
                'content_type': 'html',
                'url': _object.absolute_url,
                'published': _object.published,
                'updated': _object.published,
                'author': self.flourish.site_config['author'],
            }
            if 'author' in _object:
                entry['author'] = _object.author
            if 'updated' in _object:
                entry['updated'] = _object.updated
            _feed.add(**entry)

        return _feed.to_string()



import argparse
from hashlib import md5
from multiprocessing import Process
import os
import textwrap
import time

import boto3
from flask import Flask, send_from_directory
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler

from . import Flourish, __version__
from .examples import example_files
# FIXME every compatible instance of os.walk should use this
from .lib import relative_list_of_files_in_directory


def main():
    version = ('Flourish static site generator, version %s -- '
               'http://withaflourish.net' % __version__)
    parser = argparse.ArgumentParser(
        description=textwrap.dedent("""\
            Flourish generates websites from source data and templates.
            Version %s -- http://withaflourish.net

            Commands:
                generate        creates the output HTML
                preview         preview the generated content from your local
                                computer before uploading
                example         creates a new flourish website in the
                                current directory, with some example content
                upload          upload the generated site to an AWS S3 bucket
        """ % __version__),
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument(
        'action', nargs='?', choices=ACTIONS.keys()),
    parser.add_argument(
        '-b', '--base', default=None,
        help='Base directory that contains all others')
    parser.add_argument(
        '-s', '--source', default='source',
        help='Directory containing source files (default: %(default)s)')
    parser.add_argument(
        '-t', '--templates', default='templates',
        help='Directory containing templates (default: %(default)s)')
    parser.add_argument(
        '-o', '--output', default='output',
        help='Directory to output to (default: %(default)s)')
    parser.add_argument(
        '-p', '--port', default='3567',
        help='Port on which to bind preview webserver (default: %(default)s)')
    parser.add_argument(
        '-v', '--verbose', action='store_true',
        help='Report each URL as it is generated')
    parser.add_argument(
        '--version', action='store_true',
        help='Report the version of flourish')
    parser.add_argument(
        '--rebuild', action='store_true',
        help='Watch the source and template directories for changes, '
             'regenerating the site automatically.')
    parser.add_argument(
        '--bucket',
        help='The name of the AWS S3 bucket to upload to (with upload)')

    args = parser.parse_args()

    if args.version:
        print version
    elif args.action is None:
        parser.print_help()
    else:
        if args.base is not None:
            args.source = os.path.join(args.base, args.source)
            args.templates = os.path.join(args.base, args.templates)
            args.output = os.path.join(args.base, args.output)
        action = ACTIONS[args.action]
        action(args)


def generate(args):
    if args.rebuild:
        generate_once(args)
        generate_on_change(args)
    else:
        generate_once(args)


def generate_once(args):
    flourish = Flourish(
        source_dir=args.source,
        templates_dir=args.templates,
        output_dir=args.output,
    )
    flourish.generate_site(report=args.verbose)


def generate_on_change(args):
    class Handler(FileSystemEventHandler):
        def on_any_event(self, event):
            print '**', event.src_path, event.event_type
            generate_once(args)

    generate_once(args)
    observer = Observer()
    observer.daemon = True
    observer.schedule(Handler(), args.source, recursive=True)
    observer.schedule(Handler(), args.templates, recursive=True)
    # FIXME explodes too easily, should do source not in bg thread?
    observer.start()
    try:
        while True:
            time.sleep(0.5)
    except KeyboardInterrupt:
        observer.stop()
    observer.join()


def preview_server(args):
    if args.rebuild:
        preview_rebuildserver(args)
    else:
        preview_runserver(args)


def preview_runserver(args):
    output_dir = os.path.abspath(args.output)
    app = Flask(__name__)

    # TODO 3xx redirects
    # TODO 404 page
    # TODO 410 page
    @app.route('/')
    @app.route('/<path:path>')
    def send_file(path=''):
        filename = os.path.basename(path)
        if filename == '':
            path = '%sindex.html' % path
        root, ext = os.path.splitext(path)
        if ext == '':
            path = '%s.html' % path
        return send_from_directory(output_dir, path)

    @app.after_request
    def add_header(response):
        response.cache_control.max_age = 0
        return response

    app.run(port=3567)


def preview_rebuildserver(args):
    _server = Process(target=preview_runserver, args=(args,))
    _server.start()
    # _goc = Process(target=generate_on_change, args=(args,))
    # _goc.start()
    generate_on_change(args)


def create_example(args):
    for _dir in ['source', 'templates']:
        if not os.path.isdir(_dir):
            os.mkdir(_dir)
    for _filename in example_files:
        with open(_filename, 'w') as _example_file:
            _example_file.write(example_files[_filename])
    generate_once(argparse.Namespace(
        source='source',
        templates='templates',
        output='output',
    ))
    print 'Example site created: run "flourish --rebuild preview"'
    print 'and go to http://localhost:3567/ in your browser to see the site.'
    print 'Then click on "Welcome to your new blog" to get started.'


def create_new(args):
    # TODO new blog
    # TODO new brochureware site?
    # TODO new ...?
    pass


def upload(args):
    _bucket_name = args.bucket
    if _bucket_name is None:
        flourish = Flourish(args.source, args.templates, args.output)
        _bucket_name = flourish.site_config['bucket']

    # FIXME remove nonexistent files
    _s3 = boto3.resource('s3')
    _bucket = _s3.Bucket(_bucket_name)
    _files = relative_list_of_files_in_directory(args.output)
    _objects = dict()
    for _object in _bucket.objects.all():
        _objects.update({_object.key: _object})

    for _path in _files:
        _type = None
        _file = os.path.basename(_path)
        _s3path = _path
        if _file.endswith('.html'):
            _type = 'text/html'
            if not _file == 'index.html':
                _s3path = _path[:-5]

        _upload = True
        _md5 = '"%s"' % md5(
                open('%s/%s' % (args.output, _path), 'rb').read()
            ).hexdigest()
        try:
            if _md5 == _objects[_s3path].e_tag:
                _upload = False
        except KeyError:
            pass

        if _upload:
            print '->', _s3path
            _object_args = {
                'Key': _s3path,
                'ACL': 'public-read',
                'Body': open('%s/%s' % (args.output, _path), 'rb'),
            }
            if _type:
                _object_args['ContentType'] = _type
            _bucket.put_object(**_object_args)


ACTIONS = {
    'generate': generate,
    'preview': preview_server,
    'example': create_example,
    'new': create_new,
    'upload': upload,
    # 'shell': shell,
}

if __name__ == '__main__':
    main()
